@@ -24,7 +24,7 @@ WARNING: Rails was not able to disable referential integrity.
 This is most likely caused due to missing permissions.
 Rails needs superuser privileges to disable referential integrity.
 
-    cause: #{original_exception.try(:message)}
+    cause: #{original_exception&.message}
 
             WARNING
             raise e
